import { Content, ContentContent, ContentAccessor, QueryListOption } from '../model/contents';
import Knex from 'knex';
import {createKnexTagController} from './tag';
import { TagAccessor } from '../model/tag';

type DBTagContentRelation = {
    content_id:number,
    tag_name:string
}

class KnexContentsAccessor implements ContentAccessor{
    knex : Knex;
    tagController: TagAccessor;
    constructor(knex : Knex){
        this.knex = knex;
        this.tagController = createKnexTagController(knex);
    }
    async add(c: ContentContent){
        const {tags,additional, ...rest} = c;
        const id_lst = await this.knex.insert({
            additional:JSON.stringify(additional),
            ...rest
        }).into('contents');
        const id = id_lst[0];
        for (const it of tags) {
            this.tagController.addTag({name:it});
        }
        if(tags.length > 0){
            await this.knex.insert<DBTagContentRelation>(
                tags.map(x=>({content_id:id,tag_name:x}))
            ).into("content_tag_relation");
        }
        return id;
    };
    async del(id:number) {
        if (await this.findById(id) !== undefined){
            await this.knex.delete().from("content_tag_relation").where({content_id:id});
            await this.knex.delete().from("contents").where({id:id});
            return true;
        }
        return false;
    };
    async findById(id:number,tagload?:boolean): Promise<Content|undefined>{
        const s = await this.knex.select("*").from("contents").where({id:id});
        if(s.length === 0) return undefined;
        const first = s[0];
        let ret_tags:string[] = []
        if(tagload === true){
            const tags : DBTagContentRelation[] = await this.knex.select("*")
                .from("content_tag_relation").where({content_id:first.id});
            ret_tags = tags.map(x=>x.tag_name);
        }
        return {
            ...first,
            tags:ret_tags,
            additional: JSON.parse(first.additional || "{}"),
        };
    };
    async findList(option?:QueryListOption){
        option = option || {};
        const allow_tag = option.allow_tag || [];
        const eager_loading = typeof option.eager_loading === "undefined" || option.eager_loading;
        const limit = option.limit || 20;
        const use_offset = option.use_offset || false;
        const offset = option.offset || 0;
        const word = option.word;
        const content_type = option.content_type;
        const cursor = option.cursor;

        const buildquery = ()=>{
            let query = this.knex.select("*");
            if(allow_tag.length > 0){
                query = query.from("content_tag_relation").innerJoin("contents","content_tag_relation.content_id","contents.id");
                for(const tag of allow_tag){
                    query = query.where({tag_name:tag});
                }
            }
            else{
                query = query.from("contents");
            }
            if(word !== undefined){
                //don't worry about sql injection.
                query = query.where('title','like',`%${word}%`);
            }
            if(content_type !== undefined){
                query = query.where('content_type','=',content_type);
            }
            if(use_offset){
                query = query.offset(offset);
            }
            else{
                if(cursor !== undefined){
                    query = query.where('id','<',cursor);
                }
            }
            query = query.limit(limit);
            return query;
        }
        let query = buildquery();
        //console.log(query.toSQL());
        let result:Content[] = await query;
        for(let i of result){
            i.additional = JSON.parse((i.additional as unknown) as string);
        }
        if(eager_loading){
            let idmap: {[index:number]:Content} = {};
            for(const r of result){
                idmap[r.id] = r;
                r.tags = [];
            }
            let subquery = buildquery();
            let tagresult:{id:number,tag_name:string}[] = await this.knex.select("id","content_tag_relation.tag_name").from(subquery)
                .innerJoin("content_tag_relation","content_tag_relation.content_id","id");
            for(const {id,tag_name} of tagresult){
                idmap[id].tags.push(tag_name);
            }
        }
        return result;
    };
    async findListByBasePath(path:string):Promise<Content[]>{
        let results = await this.knex.select("*").from("contents").where({basepath:path});
        return results.map(x=>({
            ...x,
            tags:[],
            additional:JSON.parse(x.additional || "{}"),
        }));
    }
    async update(c:Partial<Content> & { id:number }){
        const {id,tags,...rest} = c;
        if (await this.findById(id) !== undefined){
            await this.knex.update(rest).where({id: id}).from("contents");
            return true;
        }
        return false;
    }
    async addTag(c: Content,tag_name:string){
        if (c.tags.includes(tag_name)) return false;
        this.tagController.addTag({name:tag_name});
        await this.knex.insert<DBTagContentRelation>({tag_name: tag_name, content_id: c.id})
            .into("content_tag_relation");
        c.tags.push(tag_name);
        return true;
    }
    async delTag(c: Content,tag_name:string){
        if (c.tags.includes(tag_name)) return false;
        await this.knex.delete().where({tag_name: tag_name,content_id: c.id}).from("content_tag_relation");
        c.tags.push(tag_name);
        return true;
    }
}
export const createKnexContentsAccessor = (knex:Knex): ContentAccessor=>{
    return new KnexContentsAccessor(knex);
}