import {sign, decode, verify} from 'jsonwebtoken';
import Koa from 'koa';
import Router from 'koa-router';
import { sendError } from './route/error_handler';
import Knex from 'knex'
import { createKnexUserController } from './db/mod';
import { request } from 'http';
import { get_setting } from './setting';
import { IUser } from './model/mod';

type PayloadInfo = {
    username:string,
    permission:string[]
}

export type UserState = {
    user?:PayloadInfo
};

const isUserState = (obj:object|string):obj is PayloadInfo =>{
    if(typeof obj ==="string") return false;
    return 'username' in obj && 'permission' in obj && (obj as {permission:unknown}).permission instanceof Array;
}

export const loginTokenName = 'access_token'

export const getAdminCookieValue = ()=>{
    const setting = get_setting();
    const secretKey = setting.jwt_secretkey;
    return sign({
        username: "admin",
        permission: [],
    },secretKey,{expiresIn:'3d'});
}

export const createLoginMiddleware = (knex: Knex)=>{
    const userController = createKnexUserController(knex);
    return async (ctx: Koa.Context,next: Koa.Next)=>{
        const setting = get_setting();
        const secretKey = setting.jwt_secretkey;
        const body = ctx.request.body;
        if(!('username' in body)||!('password' in body)){
            return sendError(400,"invalid form : username or password is not found in query.");
        }
        const username = body['username'];
        const password = body['password'];
        if(typeof username !== "string" || typeof password !== "string"){
            return sendError(400,"invalid form : username or password is not string")
        }
        if(setting.forbid_remote_admin_login && username === "admin"){
            return sendError(403,"forbid remote admin login");
        }
        const user = await userController.findUser(username);
        if(user === undefined){
            return sendError(401,"not authorized");
        }
        if(!user.password.check_password(password)){
            return sendError(401,"not authorized");
        }
        const userPermission = await user.get_permissions();
        const payload = sign({
            username: user.username,
            permission: userPermission
        },secretKey,{expiresIn:'3h'});
        ctx.cookies.set(loginTokenName,payload,{httpOnly:true, secure: !setting.localmode,sameSite:'strict'});
        ctx.body = {ok:true, username: user.username, permission: userPermission}
        console.log(`${username} logined`);
        return;
    };
};
export const LogoutMiddleware = (ctx:Koa.Context,next:Koa.Next)=>{
    ctx.cookies.set(loginTokenName,undefined);
    ctx.body = {ok:true};
    return;
}
export const UserMiddleWare = async (ctx:Koa.ParameterizedContext<UserState>,next:Koa.Next)=>{
    const secretKey = get_setting().jwt_secretkey;
    const payload = ctx.cookies.get(loginTokenName);
    if(payload == undefined){
        ctx.state['user'] = undefined;
        return await next();
    }
    const o = verify(payload,secretKey);
    if(isUserState(o)){
        ctx.state.user = o;
        return await next();
    }
    else{
        console.error("invalid token detected");
    }
}

export const getAdmin = async(knex : Knex)=>{
    const cntr = createKnexUserController(knex);
    const admin = await cntr.findUser("admin");
    if(admin === undefined){
        throw new Error("initial process failed!");//???
    }
    return admin;
}

export const isAdminFirst = (admin: IUser)=>{
    return admin.password.hash === "unchecked" && admin.password.salt === "unchecked";
}